@@ -70,9 +70,16 @@ flag peek_capable;	/* can we peek for better error recovery? */
 int mailserver_socket_temp = -1;	/* socket to free if connect timeout */ 
 
 static int timeoutcount;		/* count consecutive timeouts */
+static int idletimeout;			/* timeout occured in idle stage? */
 
 static jmp_buf	restart;
 
+int isidletimeout(void)
+/* last timeout occured in idle stage? */
+{
+    return idletimeout;
+}
+
 void set_timeout(int timeleft)
 /* reset the nonresponse-timeout */
 {
@@ -82,6 +89,8 @@ void set_timeout(int timeleft)
     if (timeleft == 0)
 	timeoutcount = 0;
 
+    idletimeout = 1;
+
     ntimeout.it_interval.tv_sec = ntimeout.it_interval.tv_usec = 0;
     ntimeout.it_value.tv_sec  = timeleft;
     ntimeout.it_value.tv_usec = 0;
@@ -89,34 +98,34 @@ void set_timeout(int timeleft)
 #endif
 }
 
-static void timeout_handler (int signal)
+static RETSIGTYPE timeout_handler (int signal)
 /* handle SIGALRM signal indicating a server timeout */
 {
-    timeoutcount++;
-    longjmp(restart, THROW_TIMEOUT);
+    if(stage != STAGE_IDLE) {
+	timeoutcount++;
+	longjmp(restart, THROW_TIMEOUT);
+    } else
+	idletimeout = 1;
 }
 
-static void sigpipe_handler (int signal)
+static RETSIGTYPE sigpipe_handler (int signal)
 /* handle SIGPIPE signal indicating a broken stream socket */
 {
     longjmp(restart, THROW_SIGPIPE);
 }
 
-/* ignore SIGALRM signal indicating a timeout during cleanup */
-static void cleanup_timeout_handler (int signal) { }
-
 #define CLEANUP_TIMEOUT 60 /* maximum timeout during cleanup */
 
 static int cleanupSockClose (int fd)
 /* close sockets in maximum CLEANUP_TIMEOUT seconds during cleanup */
 {
     int scerror;
-    void (*alrmsave)(int);
-    alrmsave = signal(SIGALRM, cleanup_timeout_handler);
+    SIGHANDLERTYPE alrmsave;
+    alrmsave = set_signal_handler(SIGALRM, null_signal_handler);
     set_timeout(CLEANUP_TIMEOUT);
     scerror = SockClose(fd);
     set_timeout(0);
-    signal(SIGALRM, alrmsave);
+    set_signal_handler(SIGALRM, alrmsave);
     return (scerror);
 }
 
@@ -321,8 +330,8 @@ static void send_size_warnings(struct query *ctl)
     if (open_warning_by_mail(ctl, (struct msgblk *)NULL))
 	return;
     stuff_warning(ctl,
-	   GT_("Subject: Fetchmail oversized-messages warning.\r\n"
-	     "\r\n"
+	   GT_("Subject: Fetchmail oversized-messages warning.\n"
+	     "\n"
 	     "The following oversized messages remain on the mail server %s:"),
 		  ctl->server.pollname);
  
@@ -339,7 +348,7 @@ static void send_size_warnings(struct query *ctl)
 	    nbr = current->val.status.mark;
 	    size = atoi(current->id);
 	    stuff_warning(ctl, 
-		    GT_("\t%d msg %d octets long skipped by fetchmail.\r\n"),
+		    GT_("\t%d msg %d octets long skipped by fetchmail.\n"),
 		    nbr, size);
 	}
 	current->val.status.num++;
@@ -420,7 +429,16 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	{
 	    if ((msgcodes[num-1] == MSGLEN_TOOLARGE) && !check_only)
 		mark_oversized(ctl, num, msgsizes[num-1]);
-	    if (outlevel > O_SILENT)
+  		/* To avoid flooding the syslog when using --keep,
+  		 * report "Skipped message" only when:
+  		 *  1) --verbose is on, or
+  		 *  2) fetchmail does not use syslog, or
+  		 *  3) the message was skipped for some other
+  		 *     reason than being old.
+  		 */
+	    if (   (outlevel >= O_VERBOSE) ||
+	           (outlevel > O_SILENT && (!run.use_syslog || msgcodes[num-1] != MSGLEN_OLD))
+	       )
 	    {
 		report_build(stdout, 
 			     GT_("skipping message %s@%s:%d (%d octets)"),
@@ -493,28 +511,33 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	    err = readheaders(mailserver_socket, len, msgsizes[num-1],
 			     ctl, num);
 	    if (err == PS_RETAINED)
-		suppress_readbody = suppress_forward = suppress_delete = retained = TRUE;
+	    {
+		suppress_forward = suppress_delete = retained = TRUE;
+		/* do not read the body only if the underlying protocol
+		 * allows the body to be fetched separately */
+		if (ctl->server.base_protocol->fetch_body)
+		    suppress_readbody = TRUE;
+	    }
 	    else if (err == PS_TRANSIENT)
+	    {
 		suppress_delete = suppress_forward = TRUE;
+		if (ctl->server.base_protocol->fetch_body)
+		    suppress_readbody = TRUE;
+	    }
 	    else if (err == PS_REFUSED)
+	    {
 		suppress_forward = TRUE;
-#if 0
-	    /* 
-	     * readheaders does not read the body when it
-	     * hits a non-header. It has been recently
-	     * fixed to return PS_TRUNCATED (properly) when
-	     * that happens, but apparently fixing that bug
-	     * opened this one here (which looks like an 
-	     * inproper fix from some ancient thinko)
-	     */
+		if (ctl->server.base_protocol->fetch_body)
+		    suppress_readbody = TRUE;
+	    }
 	    else if (err == PS_TRUNCATED)
-		suppress_readbody = TRUE;
+	    {
+		if (ctl->server.base_protocol->fetch_body)
+		    suppress_readbody = TRUE;
+		len = 0;	/* suppress body processing */
+	    }
 	    else if (err)
 		return(err);
-#else
-	    else if (err && err != PS_TRUNCATED)
-		return(err);
-#endif
 
 	    /* 
 	     * If we're using IMAP4 or something else that
@@ -644,25 +667,6 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	 * now.
 	 */
 
-	/*
-	 * Tell the UID code we've seen this.
-	 * Matthias Andree: only register the UID if we could actually
-	 * forward this mail. If we omit this !suppress_delete check,
-	 * fetchmail will never retry mail that the local listener
-	 * refused temporarily.
-	 */
-	if (ctl->newsaved && !suppress_delete)
-	{
-	    struct idlist	*sdp;
-
-	    for (sdp = ctl->newsaved; sdp; sdp = sdp->next)
-		if ((sdp->val.status.num == num) && (msgcodes[num-1] >= 0))
-		{
-		    sdp->val.status.mark = UID_SEEN;
-		    save_str(&ctl->oldsaved, sdp->id,UID_SEEN);
-		}
-	}
-
 	/* maybe we delete this message now? */
 	if (retained)
 	{
@@ -679,13 +683,32 @@ static int fetch_messages(int mailserver_socket, struct query *ctl,
 	    err = (ctl->server.base_protocol->delete)(mailserver_socket, ctl, num);
 	    if (err != 0)
 		return(err);
-#ifdef POP3_ENABLE
-	    delete_str(&ctl->newsaved, num);
-#endif /* POP3_ENABLE */
 	}
-	else if (outlevel > O_SILENT)
+	else
+	{
+	    if (   (outlevel >= O_VERBOSE) ||
+         		/* To avoid flooding the syslog when using --keep,
+         		 * report "Skipped message" only when:
+         		 *  1) --verbose is on, or
+         		 *  2) fetchmail does not use syslog, or
+         		 *  3) the message was skipped for some other
+         		 *     reason than just being old.
+         		 */
+	           (outlevel > O_SILENT && (!run.use_syslog || msgcodes[num-1] != MSGLEN_OLD))
+	       )
 	    report_complete(stdout, GT_(" not flushed\n"));
 
+	    /* maybe we mark this message as seen now? */
+	    if (ctl->server.base_protocol->mark_seen
+		&& !suppress_delete
+		&& (msgcodes[num-1] >= 0 && ctl->keep))
+	    {
+		err = (ctl->server.base_protocol->mark_seen)(mailserver_socket, ctl, num);
+		if (err != 0)
+		    return(err);
+	    }
+	}
+
 	/* perhaps this as many as we're ready to handle */
 	if (maxfetch && maxfetch <= *fetches && *fetches < count)
 	{
@@ -711,8 +734,8 @@ const int maxfetch;		/* maximum number of messages to fetch */
     int err, mailserver_socket = -1;
 #endif /* HAVE_VOLATILE */
     const char *msg;
-    void (*pipesave)(int);
-    void (*alrmsave)(int);
+    SIGHANDLERTYPE pipesave;
+    SIGHANDLERTYPE alrmsave;
 
     ctl->server.base_protocol = proto;
 
@@ -721,11 +744,11 @@ const int maxfetch;		/* maximum number of messages to fetch */
     init_transact(proto);
 
     /* set up the server-nonresponse timeout */
-    alrmsave = signal(SIGALRM, timeout_handler);
+    alrmsave = set_signal_handler(SIGALRM, timeout_handler);
     mytimeout = ctl->server.timeout;
 
     /* set up the broken-pipe timeout */
-    pipesave = signal(SIGPIPE, sigpipe_handler);
+    pipesave = set_signal_handler(SIGPIPE, sigpipe_handler);
 
     if ((js = setjmp(restart)))
     {
@@ -747,7 +770,7 @@ const int maxfetch;		/* maximum number of messages to fetch */
 	
 	if (js == THROW_SIGPIPE)
 	{
-	    signal(SIGPIPE, SIG_IGN);
+	    set_signal_handler(SIGPIPE, SIG_IGN);
 	    report(stdout,
 		   GT_("SIGPIPE thrown from an MDA or a stream socket error\n"));
 	    wait(0);
@@ -784,18 +807,18 @@ const int maxfetch;		/* maximum number of messages to fetch */
 		&& !open_warning_by_mail(ctl, (struct msgblk *)NULL))
 	    {
 		stuff_warning(ctl,
-			      GT_("Subject: fetchmail sees repeated timeouts\r\n"));
+			      GT_("Subject: fetchmail sees repeated timeouts\n"));
 		stuff_warning(ctl,
-			      GT_("Fetchmail saw more than %d timeouts while attempting to get mail from %s@%s.\r\n"), 
+			      GT_("Fetchmail saw more than %d timeouts while attempting to get mail from %s@%s.\n"), 
 			      MAX_TIMEOUTS,
 			      ctl->remotename,
 			      ctl->server.truename);
 		stuff_warning(ctl, 
-    GT_("This could mean that your mailserver is stuck, or that your SMTP\r\n" \
-    "server is wedged, or that your mailbox file on the server has been\r\n" \
-    "corrupted by a server error.  You can run `fetchmail -v -v' to\r\n" \
-    "diagnose the problem.\r\n\r\n" \
-    "Fetchmail won't poll this mailbox again until you restart it.\r\n"));
+    GT_("This could mean that your mailserver is stuck, or that your SMTP\n" \
+    "server is wedged, or that your mailbox file on the server has been\n" \
+    "corrupted by a server error.  You can run `fetchmail -v -v' to\n" \
+    "diagnose the problem.\n\n" \
+    "Fetchmail won't poll this mailbox again until you restart it.\n"));
 		close_warning_by_mail(ctl, (struct msgblk *)NULL);
 		ctl->wedged = TRUE;
 	    }
@@ -983,8 +1006,8 @@ const int maxfetch;		/* maximum number of messages to fetch */
 		if (open_warning_by_mail(ctl, (struct msgblk *)NULL) == 0)
 		{
 		    stuff_warning(ctl,
-			 GT_("Subject: Fetchmail unreachable-server warning.\r\n"
-			   "\r\n"
+			 GT_("Subject: Fetchmail unreachable-server warning.\n"
+			   "\n"
 			   "Fetchmail could not reach the mail server %s:")
 				  ctl->server.pollname);
 		    stuff_warning(ctl, errbuf, ctl->server.pollname);
@@ -1090,49 +1113,50 @@ const int maxfetch;		/* maximum number of messages to fetch */
 		     * we let the user know service is restored.
 		     */
 		    if (run.poll_interval
-			&& ctl->wehavesentauthnote
-			&& ((ctl->wehaveauthed && ++ctl->authfailcount == 10)
-			    || ++ctl->authfailcount == 3)
+			&& !ctl->wehavesentauthnote
+			&& ((ctl->wehaveauthed && ++ctl->authfailcount >= 10)
+			    || (!ctl->wehaveauthed && ++ctl->authfailcount >= 3))
 			&& !open_warning_by_mail(ctl, (struct msgblk *)NULL))
 		    {
 			ctl->wehavesentauthnote = 1;
 			stuff_warning(ctl,
-				      GT_("Subject: fetchmail authentication failed on %s@%s\r\n"),
+				      GT_("Subject: fetchmail authentication failed on %s@%s\n"),
 			    ctl->remotename, ctl->server.truename);
 			stuff_warning(ctl,
-				      GT_("Fetchmail could not get mail from %s@%s.\r\n"), 
+				      GT_("Fetchmail could not get mail from %s@%s.\n"), 
 				      ctl->remotename,
 				      ctl->server.truename);
 			if (ctl->wehaveauthed)
 			    stuff_warning(ctl, GT_("\
-The attempt to get authorization failed.\r\n\
-Since we have already succeeded in getting authorization for this\r\n\
-connection, this is probably another failure mode (such as busy server)\r\n\
-that fetchmail cannot distinguish because the server didn't send a useful\r\n\
-error message.\r\n\
-\r\n\
-However, if you HAVE changed you account details since starting the\r\n\
-fetchmail daemon, you need to stop the daemon, change your configuration\r\n\
-of fetchmail, and then restart the daemon.\r\n\
-\r\n\
-The fetchmail daemon will continue running and attempt to connect\r\n\
-at each cycle.  No future notifications will be sent until service\r\n\
+The attempt to get authorization failed.\n\
+Since we have already succeeded in getting authorization for this\n\
+connection, this is probably another failure mode (such as busy server)\n\
+that fetchmail cannot distinguish because the server didn't send a useful\n\
+error message.\n\
+\n\
+However, if you HAVE changed your account details since starting the\n\
+fetchmail daemon, you need to stop the daemon, change your configuration\n\
+of fetchmail, and then restart the daemon.\n\
+\n\
+The fetchmail daemon will continue running and attempt to connect\n\
+at each cycle.  No future notifications will be sent until service\n\
 is restored."));
 			else
 			    stuff_warning(ctl, GT_("\
-The attempt to get authorization failed.\r\n\
-This probably means your password is invalid, but some servers have\r\n\
-other failure modes that fetchmail cannot distinguish from this\r\n\
-because they don't send useful error messages on login failure.\r\n\
-\r\n\
-The fetchmail daemon will continue running and attempt to connect\r\n\
-at each cycle.  No future notifications will be sent until service\r\n\
+The attempt to get authorization failed.\n\
+This probably means your password is invalid, but some servers have\n\
+other failure modes that fetchmail cannot distinguish from this\n\
+because they don't send useful error messages on login failure.\n\
+\n\
+The fetchmail daemon will continue running and attempt to connect\n\
+at each cycle.  No future notifications will be sent until service\n\
 is restored."));
 			close_warning_by_mail(ctl, (struct msgblk *)NULL);
 		    }
 		}
 		else if (err == PS_REPOLL)
 		{
+		  if (outlevel >= O_VERBOSE)
 		    report(stderr, GT_("Repoll immediately on %s@%s\n"),
 			   ctl->remotename,
 			   ctl->server.truename);
@@ -1168,14 +1192,14 @@ is restored."));
 		    if (!open_warning_by_mail(ctl, (struct msgblk *)NULL))
 		    {
 			stuff_warning(ctl,
-			      GT_("Subject: fetchmail authentication OK on %s@%s\r\n"),
+			      GT_("Subject: fetchmail authentication OK on %s@%s\n"),
 				      ctl->remotename, ctl->server.truename);
 			stuff_warning(ctl,
-			      GT_("Fetchmail was able to log into %s@%s.\r\n"), 
+			      GT_("Fetchmail was able to log into %s@%s.\n"), 
 				      ctl->remotename,
 				      ctl->server.truename);
 			stuff_warning(ctl, 
-				      GT_("Service has been restored.\r\n"));
+				      GT_("Service has been restored.\n"));
 			close_warning_by_mail(ctl, (struct msgblk *)NULL);
 		    
 		    }
@@ -1485,8 +1509,9 @@ closeUp:
 	    err = PS_SYNTAX;
     }
 
-    signal(SIGALRM, alrmsave);
-    signal(SIGPIPE, pipesave);
+    set_timeout(0); /* cancel any pending alarm */
+    set_signal_handler(SIGALRM, alrmsave);
+    set_signal_handler(SIGPIPE, pipesave);
     return(err);
 }
 
