@@ -31,6 +31,7 @@ extern char *strstr();	/* needed on sysV68 R3V7.1. */
 #define IMAP4rev1	1	/* IMAP4 rev 1, RFC2060 */
 
 static int count = 0, recentcount = 0, unseen = 0, deletions = 0;
+static unsigned int startcount = 1;
 static int expunged, expunge_period, saved_timeout = 0;
 static int imap_version, preauth;
 static flag do_idle;
@@ -56,7 +57,10 @@ static int imap_ok(int sock, char *argbuf)
 
 	/* interpret untagged status responses */
 	if (strstr(buf, "* CAPABILITY"))
+	{
 	    strncpy(capabilities, buf + 12, sizeof(capabilities));
+	    capabilities[sizeof(capabilities)-1] = '\0';
+	}
 	else if (strstr(buf, "EXISTS"))
 	{
 	    count = atoi(buf+2);
@@ -91,7 +95,8 @@ static int imap_ok(int sock, char *argbuf)
 		stage = STAGE_FETCH;
 	    }
 	}
-	else if (strstr(buf, "RECENT"))
+	/* a space is required to avoid confusion with the \Recent flag */
+	else if (strstr(buf, " RECENT"))
 	{
 	    recentcount = atoi(buf+2);
 	}
@@ -132,15 +137,15 @@ static int imap_ok(int sock, char *argbuf)
 	while (isspace(*cp))
 	    cp++;
 
-        if (strncmp(cp, "OK", 2) == 0)
+        if (strncasecmp(cp, "OK", 2) == 0)
 	{
 	    if (argbuf)
 		strcpy(argbuf, cp);
 	    return(PS_SUCCESS);
 	}
-	else if (strncmp(cp, "BAD", 3) == 0)
+	else if (strncasecmp(cp, "BAD", 3) == 0)
 	    return(PS_ERROR);
-	else if (strncmp(cp, "NO", 2) == 0)
+	else if (strncasecmp(cp, "NO", 2) == 0)
 	{
 	    if (stage == STAGE_GETAUTH) 
 		return(PS_AUTHFAIL);	/* RFC2060, 6.2.2 */
@@ -246,10 +251,10 @@ static int imap_canonicalize(char *result, char *raw, int maxlen)
     return(i);
 }
 
-static int imap_getauth(int sock, struct query *ctl, char *greeting)
-/* apply for connection authorization */
+static void capa_probe(int sock, struct query *ctl)
+/* set capability variables from a CAPA probe */
 {
-    int ok = 0;
+    int	ok;
 
     /* probe to see if we're running IMAP4 and can use RFC822.PEEK */
     capabilities[0] = '\0';
@@ -282,19 +287,6 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
 	if (outlevel >= O_DEBUG)
 	    report(stdout, GT_("Protocol identified as IMAP2 or IMAP2BIS\n"));
     }
-    else
-	return(ok);
-
-    peek_capable = (imap_version >= IMAP4);
-
-    /* 
-     * Assumption: expunges are cheap, so we want to do them
-     * after every message unless user said otherwise.
-     */
-    if (NUM_SPECIFIED(ctl->expunge))
-	expunge_period = NUM_VALUE_OUT(ctl->expunge);
-    else
-	expunge_period = 1;
 
     /* 
      * Handle idling.  We depend on coming through here on startup
@@ -306,6 +298,17 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
 	if (outlevel >= O_VERBOSE)
 	    report(stdout, GT_("will idle after poll\n"));
     }
+}
+
+static int imap_getauth(int sock, struct query *ctl, char *greeting)
+/* apply for connection authorization */
+{
+    int ok = 0;
+#ifdef SSL_ENABLE
+    flag did_stls = FALSE;
+#endif /* SSL_ENABLE */
+
+    capa_probe(sock, ctl);
 
     /* 
      * If either (a) we saw a PREAUTH token in the greeting, or
@@ -316,7 +319,6 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
         preauth = FALSE;  /* reset for the next session */
         return(PS_SUCCESS);
     }
-
     /*
      * Time to authenticate the user.
      * Try the protocol variants that don't require passwords first.
@@ -357,26 +359,58 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
 #endif /* KERBEROS_V4 */
 
 #ifdef SSL_ENABLE
-    if ((ctl->server.authenticate == A_ANY)
+    if ((!ctl->sslproto || !strcmp(ctl->sslproto,"tls1"))
+        && !ctl->use_ssl
         && strstr(capabilities, "STARTTLS"))
     {
            char *realhost;
 
            realhost = ctl->server.via ? ctl->server.via : ctl->server.pollname;
-           gen_transact(sock, "STARTTLS");
+           ok = gen_transact(sock, "STARTTLS");
 
            /* We use "tls1" instead of ctl->sslproto, as we want STARTTLS,
             * not other SSL protocols
             */
-           if (SSLOpen(sock,ctl->sslcert,ctl->sslkey,"tls1",ctl->sslcertck, ctl->sslcertpath,ctl->sslfingerprint,realhost,ctl->server.pollname) == -1)
+           if (ok == PS_SUCCESS &&
+	       SSLOpen(sock,ctl->sslcert,ctl->sslkey,"tls1",ctl->sslcertck, ctl->sslcertpath,ctl->sslfingerprint,realhost,ctl->server.pollname) == -1)
            {
+	       if (!ctl->sslproto && !ctl->wehaveauthed)
+	       {
+		   ctl->sslproto = xstrdup("");
+		   /* repoll immediately */
+		   return(PS_REPOLL);
+	       }
                report(stderr,
                       GT_("SSL connection failed.\n"));
                return(PS_AUTHFAIL);
            }
+	   did_stls = TRUE;
+
+	   /*
+	    * RFC 2595 says this:
+	    *
+	    * "Once TLS has been started, the client MUST discard cached
+	    * information about server capabilities and SHOULD re-issue the
+	    * CAPABILITY command.  This is necessary to protect against
+	    * man-in-the-middle attacks which alter the capabilities list prior
+	    * to STARTTLS.  The server MAY advertise different capabilities
+	    * after STARTTLS."
+	    */
+	   capa_probe(sock, ctl);
     }
 #endif /* SSL_ENABLE */
 
+    peek_capable = (imap_version >= IMAP4);
+
+    /* 
+     * Assumption: expunges are cheap, so we want to do them
+     * after every message unless user said otherwise.
+     */
+    if (NUM_SPECIFIED(ctl->expunge))
+	expunge_period = NUM_VALUE_OUT(ctl->expunge);
+    else
+	expunge_period = 1;
+
     /*
      * No such luck.  OK, now try the variants that mask your password
      * in a challenge-response.
@@ -449,10 +483,15 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
     }
 #endif /* __UNUSED__ */
 
-    /* we're stuck with sending the password en clair */
-    if ((ctl->server.authenticate == A_ANY 
- 	 || ctl->server.authenticate == A_PASSWORD) 
- 	&& !strstr (capabilities, "LOGINDISABLED"))
+    /* 
+     * We're stuck with sending the password en clair.
+     * The reason for this odd-looking logic is that some
+     * servers return LOGINDISABLED even though login 
+     * actually works.  So arrange things in such a way that
+     * setting auth passwd makes it ignore this capability.
+     */
+    if((ctl->server.authenticate==A_ANY&&!strstr(capabilities,"LOGINDISABLED"))
+	|| ctl->server.authenticate == A_PASSWORD)
     {
 	/* these sizes guarantee no buffer overflow */
 	char	remotename[NAMELEN*2+1], password[PASSWORDLEN*2+1];
@@ -463,6 +502,16 @@ static int imap_getauth(int sock, struct query *ctl, char *greeting)
 	strcpy(shroud, password);
 	ok = gen_transact(sock, "LOGIN \"%s\" \"%s\"", remotename, password);
 	shroud[0] = '\0';
+#ifdef SSL_ENABLE
+	/* this is for servers which claim to support TLS, but actually
+	 * don't! */
+	if (did_stls && ok == PS_SOCKET && !ctl->sslproto && !ctl->wehaveauthed)
+	{
+	    ctl->sslproto = xstrdup("");
+	    /* repoll immediately */
+	    ok = PS_REPOLL;
+	}
+#endif
 	if (ok)
 	{
 	    /* SASL cancellation of authentication */
@@ -498,11 +547,33 @@ static int internal_expunge(int sock)
 static int imap_idle(int sock)
 /* start an RFC2177 IDLE */
 {
+    int ok;
+
+    /* special timeout to terminate the IDLE and re-issue it
+     * at least every 28 minutes:
+     * (the server may have an inactivity timeout) */
     stage = STAGE_IDLE;
     saved_timeout = mytimeout;
-    mytimeout = 0;
+    mytimeout = 1680; /* 28 min */
 
-    return (gen_transact(sock, "IDLE"));
+    /* enter IDLE mode */
+    ok = gen_transact(sock, "IDLE");
+
+    if(ok == PS_IDLETIMEOUT) {
+	/* send "DONE" continuation */
+	SockWrite(sock, "DONE\r\n", 6);
+	if (outlevel >= O_MONITOR)
+	    report(stdout, "IMAP> DONE\n");
+
+	/* restore normal timeout value */
+	mytimeout = saved_timeout;
+	stage = STAGE_FETCH;
+
+	/* get OK IDLE message */
+	return imap_ok(sock, NULL);
+    } else
+	/* not idle timeout */
+	return ok;
 }
 
 static int imap_getrange(int sock, 
@@ -541,6 +612,10 @@ static int imap_getrange(int sock,
 	 * for new mail.
 	 */
 
+	/* some servers do not report RECENT after an EXPUNGE. this check
+	 * forces an incorrect recentcount to be ignored. */
+	if (recentcount > count)
+	    recentcount = 0;
 	/* this is a while loop because imap_idle() might return on other
 	 * mailbox changes also */
 	while (recentcount == 0 && do_idle) {
@@ -602,6 +677,7 @@ static int imap_getrange(int sock,
 
     *countp = count;
     recentcount = 0;
+    startcount = 1;
 
     /* OK, now get a count of unseen messages and their indices */
     if (!ctl->fetchall && count > 0)
@@ -625,6 +701,10 @@ static int imap_getrange(int sock,
 		char	*ep;
 
 		cp += 8;	/* skip "* SEARCH" */
+		/* startcount is higher than count so that if there are no
+		 * unseen messages, imap_getsizes() will not need to do
+		 * anything! */
+		startcount = count + 1;
 
 		while (*cp && unseen < count)
 		{
@@ -633,24 +713,29 @@ static int imap_getrange(int sock,
 			cp++;
 		    if (*cp) 
 		    {
+			unsigned int um;
 			/*
 			 * Message numbers are between 1 and 2^32 inclusive,
 			 * so unsigned int is large enough.
 			 */
-			unseen_messages[unseen]=(unsigned int)strtol(cp,&ep,10);
-
-			if (outlevel >= O_DEBUG)
-			    report(stdout, 
-				   GT_("%u is unseen\n"), 
-				   unseen_messages[unseen]);
-		
-			unseen++;
+			um=(unsigned int)strtol(cp,&ep,10);
+			if (um <= count)
+			{
+			    unseen_messages[unseen++] = um;
+			    if (outlevel >= O_DEBUG)
+				report(stdout, GT_("%u is unseen\n"), um);
+			    if (startcount > um)
+				startcount = um;
+			}
 			cp = ep;
 		    }
 		}
 	    }
 	} while
 	    (tag[0] != '\0' && strncmp(buf, tag, strlen(tag)));
+
+	if (outlevel >= O_DEBUG && unseen > 0)
+	    report(stdout, GT_("%u is first unseen\n"), startcount);
     } else
 	unseen = -1;
 
@@ -701,23 +786,34 @@ static int imap_getsizes(int sock, int count, int *sizes)
      * on the fact that the sizes array has been preinitialized with a
      * known-bad size value.
      */
-    if (count == 1)
-	gen_send(sock, "FETCH 1 RFC822.SIZE");
-    else
-	gen_send(sock, "FETCH 1:%d RFC822.SIZE", count);
+    /* if fetchall is specified, startcount is 1;
+     * else if there is new mail, startcount is first unseen message;
+     * else startcount is greater than count.
+     */
+    if (count == startcount)
+	gen_send(sock, "FETCH %d RFC822.SIZE", count);
+    else if (count > startcount)
+	gen_send(sock, "FETCH %d:%d RFC822.SIZE", startcount, count);
+    else /* no unseen messages! */
+	return(PS_SUCCESS);
     for (;;)
     {
 	unsigned int num, size;
 	int ok;
+	char *cp;
 
 	if ((ok = gen_recv(sock, buf, sizeof(buf))))
 	    return(ok);
+	/* we want response matching to be case-insensitive */
+	for (cp = buf; *cp; cp++)
+	    *cp = toupper(*cp);
 	/* an untagged NO means that a message was not readable */
-	else if (strstr(buf, "* NO"))
+	if (strstr(buf, "* NO"))
 	    ;
 	else if (strstr(buf, "OK") || strstr(buf, "NO"))
 	    break;
-	else if (sscanf(buf, "* %u FETCH (RFC822.SIZE %u)", &num, &size) == 2) {
+	else if (sscanf(buf, "* %u FETCH (RFC822.SIZE %u)", &num, &size) == 2) 
+	{
 	    if (num > 0 && num <= count)
 	        sizes[num - 1] = size;
 	    else
@@ -823,11 +919,6 @@ static int imap_fetch_body(int sock, struct query *ctl, int number, int *lenp)
      * craps out during the message, it will still be marked `unseen' on
      * the server.
      *
-     * However...*don't* do this if we're using keep to suppress deletion!
-     * In that case, marking the seen flag is the only way to prevent the
-     * message from being re-fetched on subsequent runs (and according
-     * to RFC2060 p.43 this fetch should set Seen as a side effect).
-     *
      * According to RFC2060, and Mark Crispin the IMAP maintainer,
      * FETCH %d BODY[TEXT] and RFC822.TEXT are "functionally 
      * equivalent".  However, we know of at least one server that
@@ -845,17 +936,11 @@ static int imap_fetch_body(int sock, struct query *ctl, int number, int *lenp)
     switch (imap_version)
     {
     case IMAP4rev1:	/* RFC 2060 */
-	if (!ctl->keep)
-	    gen_send(sock, "FETCH %d BODY.PEEK[TEXT]", number);
-	else
-	    gen_send(sock, "FETCH %d BODY[TEXT]", number);
+	gen_send(sock, "FETCH %d BODY.PEEK[TEXT]", number);
 	break;
 
     case IMAP4:		/* RFC 1730 */
-	if (!ctl->keep)
-	    gen_send(sock, "FETCH %d RFC822.TEXT.PEEK", number);
-	else
-	    gen_send(sock, "FETCH %d RFC822.TEXT", number);
+	gen_send(sock, "FETCH %d RFC822.TEXT.PEEK", number);
 	break;
 
     default:		/* RFC 1176 */
@@ -905,26 +990,6 @@ static int imap_trail(int sock, struct query *ctl, int number)
 	/* UW IMAP returns "OK FETCH", Cyrus returns "OK Completed" */
 	if (strstr(buf, "OK"))
 	    break;
-
-#ifdef __UNUSED__
-	/*
-	 * Any IMAP server that fails to set Seen on a BODY[TEXT]
-	 * fetch violates RFC2060 p.43 (top).  This becomes an issue
-	 * when keep is on, because seen messages aren't deleted and
-	 * get refetched on each poll.  As a workaround, if keep is on
-	 * we can set the Seen flag explicitly.
-	 *
-	 * This code isn't used yet because we don't know of any IMAP
-	 * servers broken in this way.
-	 */
-	if (ctl->keep)
-	    if ((ok = gen_transact(sock,
-			imap_version == IMAP4 
-				? "STORE %d +FLAGS.SILENT (\\Seen)"
-				: "STORE %d +FLAGS (\\Seen)", 
-			number)))
-		return(ok);
-#endif /* __UNUSED__ */
     }
 
     return(PS_SUCCESS);
@@ -969,6 +1034,16 @@ static int imap_delete(int sock, struct query *ctl, int number)
     return(PS_SUCCESS);
 }
 
+static int imap_mark_seen(int sock, struct query *ctl, int number)
+/* mark the given message as seen */
+{
+    return(gen_transact(sock,
+	imap_version == IMAP4
+	? "STORE %d +FLAGS.SILENT (\\Seen)"
+	: "STORE %d +FLAGS (\\Seen)",
+	number));
+}
+
 static int imap_logout(int sock, struct query *ctl)
 /* send logout command */
 {
@@ -1006,6 +1081,7 @@ const static struct method imap =
     imap_fetch_body,	/* request given message body */
     imap_trail,		/* eat message trailer */
     imap_delete,	/* delete the message */
+    imap_mark_seen,	/* how to mark a message as seen */
     imap_logout,	/* expunge and exit */
     TRUE,		/* yes, we can re-poll */
 };
