@@ -9,11 +9,24 @@
 #ifdef HAVE_RES_SEARCH
 #include <stdio.h>
 #include <string.h>
+#ifdef HAVE_NET_SOCKET_H
+#include <net/socket.h>
+#endif
 #include <netdb.h>
 #include <sys/types.h>
 #include <netinet/in.h>
+
+#ifdef __BEOS__
+#include "beos/beos_nameser.h"
+#endif
+
+#ifdef HAVE_ARPA_NAMESER_H
 #include <arpa/nameser.h>
+#endif
+#ifdef HAVE_RESOLV_H
 #include <resolv.h>
+#endif
+
 #include "mx.h"
 
 /*
@@ -55,6 +68,8 @@ struct mxentry *getmxrecords(const char *name)
     n = res_search(name, C_IN,T_MX, (unsigned char *)&answer, sizeof(answer));
     if (n == -1)
 	return((struct mxentry *)NULL);
+    if (n > sizeof(answer))
+	n = sizeof(answer);    	
 
     hp = (HEADER *)&answer;
     cp = answer + HFIXEDSZ;
