@@ -188,6 +188,8 @@ static char *parse_received(struct query *ctl, char *bufp)
     char *base, *ok = (char *)NULL;
     static char rbuf[HOSTLEN + USERNAMELEN + 4]; 
 
+#define RBUF_WRITE(value) if (tp < rbuf+sizeof(rbuf)-1) *tp++=value
+
     /*
      * Try to extract the real envelope addressee.  We look here
      * specifically for the mailserver's Received line.
@@ -215,7 +217,7 @@ static char *parse_received(struct query *ctl, char *bufp)
 		continue;
 	    tp = rbuf;
 	    for (; !isspace(*sp); sp++)
-		*tp++ = *sp;
+		RBUF_WRITE(*sp);
 	    *tp = '\0';
 
 	    /* look for valid address */
@@ -263,7 +265,7 @@ static char *parse_received(struct query *ctl, char *bufp)
 		    continue;
 		tp = rbuf;
 		for (; !isspace(*sp); sp++)
-		    *tp++ = *sp;
+		    RBUF_WRITE(*sp);
 		*tp = '\0';
 
 		if (strchr(rbuf, '@'))
@@ -281,8 +283,8 @@ static char *parse_received(struct query *ctl, char *bufp)
 	    for (sp = ok + 4; isspace(*sp); sp++)
 		continue;
 	    tp = rbuf;
-	    *tp++ = ':';	/* Here is the hack.  This is to be friends */
-	    *tp++ = ' ';	/* with nxtaddr()... */
+	    RBUF_WRITE(':');	/* Here is the hack.  This is to be friends */
+	    RBUF_WRITE(' ');	/* with nxtaddr()... */
 	    if (*sp == '<')
 	    {
 		want_gt = TRUE;
@@ -295,14 +297,17 @@ static char *parse_received(struct query *ctl, char *bufp)
                    && (want_gt ? (*sp != '>') : !isspace(*sp))
                    && *sp != ';')
 		if (!isspace(*sp))
-		    *tp++ = *sp++;
+		{
+		    RBUF_WRITE(*sp);
+		    sp++;
+		}    
 		else
 		{
 		    /* uh oh -- whitespace here can't be right! */
 		    ok = (char *)NULL;
 		    break;
 		}
-	    *tp++ = '\n';
+	    RBUF_WRITE('\n');
 	    *tp = '\0';
 	    if (strlen(rbuf) <= 3)	/* apparently nothing has been found */
 		ok = NULL;
@@ -334,6 +339,14 @@ static int sizeticker;
 
 #define EMPTYLINE(s)	((s)[0] == '\r' && (s)[1] == '\n' && (s)[2] == '\0')
 
+static int end_of_header (const char *s)
+/* accept "\r*\n" as EOH in order to be bulletproof against broken survers */
+{
+    while (s[0] == '\r')
+	s++;
+    return (s[0] == '\n' && s[1] == '\0');
+}
+
 int readheaders(int sock,
 		       long fetchlen,
 		       long reallen,
@@ -360,12 +373,14 @@ int readheaders(int sock,
     int			from_offs, reply_to_offs, resent_from_offs;
     int			app_from_offs, sender_offs, resent_sender_offs;
     int			env_offs;
-    char		*received_for, *rcv, *cp, *delivered_to;
+    char		*received_for, *rcv, *cp;
+    static char		*delivered_to = NULL;
     int 		n, linelen, oldlen, ch, remaining, skipcount;
     struct idlist 	*idp;
     flag		no_local_matches = FALSE;
     flag		headers_ok, has_nuls;
     int			olderrs, good_addresses, bad_addresses;
+    int			retain_mail = 0;
 
     sizeticker = 0;
     has_nuls = headers_ok = FALSE;
@@ -383,6 +398,9 @@ int readheaders(int sock,
      */
     if (msgblk.headers)
        free(msgblk.headers);
+    free_str_list(&msgblk.recipients);
+    if (delivered_to)
+	free(delivered_to);
 
     /* initially, no message ID */
     if (ctl->thisid)
@@ -466,7 +484,7 @@ int readheaders(int sock,
 		}
 
 	    /* check for end of headers */
-	    if (EMPTYLINE(line))
+	    if (end_of_header(line))
 	    {
 		headers_ok = TRUE;
 		has_nuls = (linelen != strlen(line));
@@ -496,10 +514,12 @@ int readheaders(int sock,
 	     */
 	    if (!isspace(line[0]) && !strchr(line, ':'))
 	    {
-		headers_ok = FALSE;
-		has_nuls = (linelen != strlen(line));
-		free(line);
-		goto process_headers;
+		/* the headers_ok flag cannot be set to FALSE here as the
+		 * delimiter has not occurred yet. */
+		if (outlevel > O_SILENT)
+		    report(stdout,
+			   GT_("incorrect header line found while scanning headers\n"));
+		retain_mail = 1;
 	    }
 
 	    /* check for RFC822 continuations */
@@ -553,9 +573,8 @@ int readheaders(int sock,
 #endif /* POP2_ENABLE */
 	    if (num == 1 && !strncasecmp(line, "X-IMAP:", 7)) {
 		free(line);
-		free(msgblk.headers);
-		msgblk.headers = NULL;
-		return(PS_RETAINED);
+		retain_mail = 1;
+		continue;
 	    }
 
 	/*
@@ -660,7 +679,8 @@ int readheaders(int sock,
 	 */
 	if (!strncasecmp("Return-Path:", line, 12) && (cp = nxtaddr(line)))
 	{
-	    strcpy(msgblk.return_path, cp);
+	    strncpy(msgblk.return_path, cp, sizeof(msgblk.return_path));
+	    msgblk.return_path[sizeof(msgblk.return_path)-1] = '\0';
 	    if (!ctl->mda) {
 		free(line);
 		continue;
@@ -792,6 +812,13 @@ int readheaders(int sock,
     }
 
  process_headers:    
+
+    if (retain_mail)
+    {
+	free(msgblk.headers);
+	msgblk.headers = NULL;
+	return(PS_RETAINED);
+    }
     /*
      * When mail delivered to a multidrop mailbox on the server is
      * addressed to multiple people on the client machine, there will
@@ -837,17 +864,6 @@ int readheaders(int sock,
 	}
     }
 
-    /*
-     * We want to detect this early in case there are so few headers that the
-     * dispatch logic barfs.
-     */
-    if (!headers_ok)
-    {
-	if (outlevel > O_SILENT)
-	    report(stdout,
-		   GT_("message delimiter found while scanning headers\n"));
-    }
-
     /*
      * Hack time.  If the first line of the message was blank, with no headers
      * (this happens occasionally due to bad gatewaying software) cons up
@@ -911,8 +927,10 @@ int readheaders(int sock,
 	else if (reply_to_offs >= 0 && (ap = nxtaddr(msgblk.headers + reply_to_offs)));
 	else if (app_from_offs >= 0 && (ap = nxtaddr(msgblk.headers + app_from_offs)));
 	/* multi-line MAIL FROM addresses confuse SMTP terribly */
-	if (ap && !strchr(ap, '\n')) 
-	    strcpy(msgblk.return_path, ap);
+	if (ap && !strchr(ap, '\n')) {
+	    strncpy(msgblk.return_path, ap, sizeof(msgblk.return_path));
+	    msgblk.return_path[sizeof(msgblk.return_path)-1] = '\0';
+	}
     }
 
     /* cons up a list of local recipients */
@@ -938,6 +956,7 @@ int readheaders(int sock,
    {
 	    find_server_names(delivered_to, ctl, &msgblk.recipients);
        free(delivered_to);
+       delivered_to = NULL;
    }
 	else if (received_for)
 	    /*
@@ -1161,22 +1180,39 @@ int readheaders(int sock,
 		for (idp = msgblk.recipients; idp; idp = idp->next)
 		    if (idp->val.status.mark == XMIT_REJECT)
 			break;
-		sprintf(errhd+strlen(errhd), GT_("recipient address %s didn't match any local name"), idp->id);
+#ifdef HAVE_SNPRINTF
+		snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd),
+#else
+		sprintf(errhd+strlen(errhd),
+#endif /* HAVE_SNPRINTF */
+			GT_("recipient address %s didn't match any local name"), idp->id);
 	    }
 	}
 
 	if (has_nuls)
 	{
 	    if (errhd[sizeof("X-Fetchmail-Warning: ")])
+#ifdef HAVE_SNPRINTF
+		snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd), "; ");
+	    snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd),
+#else
 		strcat(errhd, "; ");
-	    strcat(errhd, GT_("message has embedded NULs"));
+	    strcat(errhd,
+#endif /* HAVE_SNPRINTF */
+			GT_("message has embedded NULs"));
 	}
 
 	if (bad_addresses)
 	{
 	    if (errhd[sizeof("X-Fetchmail-Warning: ")])
+#ifdef HAVE_SNPRINTF
+		snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd), "; ");
+	    snprintf(errhd+strlen(errhd), sizeof(errhd)-strlen(errhd),
+#else
 		strcat(errhd, "; ");
-	    strcat(errhd, GT_("SMTP listener rejected local recipient addresses: "));
+	    strcat(errhd,
+#endif /* HAVE_SNPRINTF */
+			GT_("SMTP listener rejected local recipient addresses: "));
 	    errlen = strlen(errhd);
 	    for (idp = msgblk.recipients; idp; idp = idp->next)
 		if (idp->val.status.mark == XMIT_RCPTBAD)
@@ -1207,9 +1243,15 @@ int readheaders(int sock,
     *cp++ = '\0';
     stuffline(ctl, buf);
 
-/*    free(msgblk.headers); */
-    free_str_list(&msgblk.recipients);
-    return(headers_ok ? PS_SUCCESS : PS_TRUNCATED);
+    if (!headers_ok)
+    {
+	if (outlevel > O_SILENT)
+	    report(stdout,
+		   GT_("message delimiter found while scanning headers\n"));
+	return(PS_TRUNCATED);
+    }
+
+    return(PS_SUCCESS);
 }
 
 int readbody(int sock, struct query *ctl, flag forward, int len)
@@ -1404,7 +1446,10 @@ int size;	/* length of buffer */
     {
 	set_timeout(0);
 	phase = oldphase;
-	return(PS_SOCKET);
+	if(isidletimeout())
+	  return(PS_IDLETIMEOUT);
+	else
+	  return(PS_SOCKET);
     }
     else
     {
