@@ -6,6 +6,7 @@
  */
 
 #include "config.h"
+#include <sys/types.h>
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
@@ -22,28 +23,28 @@
 #endif
 
 XMALLOCTYPE *
-xmalloc (int n)
+xmalloc (size_t n)
 {
     XMALLOCTYPE *p;
 
     p = (XMALLOCTYPE *) malloc(n);
     if (p == (XMALLOCTYPE *) 0)
     {
-	report(stderr, _("malloc failed\n"));
+	report(stderr, GT_("malloc failed\n"));
 	exit(PS_UNDEFINED);
     }
     return(p);
 }
 
 XMALLOCTYPE *
-xrealloc (XMALLOCTYPE *p, int n)
+xrealloc (XMALLOCTYPE *p, size_t n)
 {
     if (p == 0)
 	return xmalloc (n);
     p = (XMALLOCTYPE *) realloc(p, n);
     if (p == (XMALLOCTYPE *) 0)
     {
-	report(stderr, _("realloc failed\n"));
+	report(stderr, GT_("realloc failed\n"));
 	exit(PS_UNDEFINED);
     }
     return p;
@@ -57,4 +58,14 @@ char *xstrdup(const char *s)
     return p;
 }
 
+#if !defined(HAVE_STRDUP)
+char *strdup(const char *s)
+{
+    char *p;
+    p = (char *) malloc(strlen(s)+1);
+    strcpy(p,s);
+    return p;
+}
+#endif /* !HAVE_STRDUP */
+
 /* xmalloc.c ends here */
